# ***************************************************************************
# Copyright 2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import os
import re
try:
    import simplejson
except ImportError:
    import json as simplejson
import sys
import tempfile
from subprocess import PIPE
from subprocess import Popen

from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.common.arguments import ScpFileUrlArgument
from cli.utils.error_code import ErrorCode as ec
from cli.utils import rest_utils
from cli.utils.error_utils import get_message
from cli.utils.file_utils import delete_file_if_exists
from cli.utils.file_utils import delete_filestore_file
from cli.utils.file_utils import get_file_store
from cli.utils.file_utils import get_tmp_store
from cli.utils.file_utils import move_file_to_filestore
from cli.utils.node_utils import get_node_type
from cli.utils.node_utils import is_appliance, is_windows
from cli.utils.rest_utils import local_node_request
from compat23 import compat23_httplib
from compat23.compat23_utils import compat23_string_types
from compat23.compat23_utils import compat23_urlparse


lg = nsx_logging.getLogger("cli.commands.common.filestore")

_PEXPECT_TIMEOUT = 120 # secs. Default ssh/scp operation's timeout: 60 secs
_SIG_HOST_KEY_ERROR = "Host key verification failed"
_SIG_NO_SPACE = "No space left on device"
_PERMISSION_DENIED = "Permission denied"
_AUTHENTICATION_FAIL = "authentication failure"
_INVALID_PATH = "No such file or directory"

_PROGRESS_STATUS_PATTERN = "(\s+)([0-9]+%)(\s+.*)"

_FILE_STORE_DIR = "/var/vmware/nsx/file-store"

_FINGERPRINT_RE = re.compile("^.+ key fingerprint is ([0-9a-zA-Z:+/]+)\..*$")


is_py3 = lambda: sys.version_info >= (3,)


def _is_pexpect_available():
    try:
        import pexpect
    except ImportError:
        return False
    return True


get_compat23_input = lambda: input if is_py3() else raw_input


def unpack_scp_cmd_values(cmd):
    """ Returns: dict(port, server, source_path, dest_path, user)"""
    data = {}
    cmd = cmd.split()
    for index in range(len(cmd)):
        if cmd[index] == "-P":
            data['port'] = int(cmd[index+1])
            data['source'] = cmd[index+2]
            dst = cmd[index+3]  # "username@hostname:path
            data['username'], path = dst.split("@")
            data['server'], data['dest'] = path.split(":")
    return data


def get_stored_thumbprint(server):
    """Retrieves the thumbprint info from know_hosts file
        returns None incase of no hosts found"""
    user_home = os.path.expanduser("~")
    known_hosts = os.path.join(user_home, ".ssh", "known_hosts")
    cmd = ["ssh-keygen", "-l", "-F", server, "-f", known_hosts]
    """
    > ssh-keygen -l -F 10.173.72.127 -f /home/admin/.ssh/known_hosts
        # Host 10.173.72.127 found: line 1
        10.173.72.127 ECDSA SHA256:aVu1Fkw58g9CXHfiuQMfhC9YCBuhvk+3M5KObh/5WtY
    """
    po = Popen(cmd, stdout=PIPE, stderr=PIPE)
    out, err = po.communicate()
    if is_py3():
        if out is not None:
            out = out.decode("utf8")
        if err is not None:
            err = err.decode("utf8")
    rc = po.returncode
    if rc == 0:
        lines = out.split("\n")
        for line in lines:
            if line and not line.startswith("#"):
                items = line.strip().split()
                if len(items) == 3 and items[0] == server:
                    return items[-1]
    lg.warn("Failed to retrieve thumbprint from known_hosts file:\
            [rc=%s, out='%s', err:'%s']" % (rc, out, err))


def _scp_file_api(data, api):
    """Transfer file to remote server using api"""
    request = {
        "port": data['port'],
        "server": data['server'],
        "uri": data['dest'],
        "protocol": {
            "name": "scp",
            "ssh_fingerprint": data['thumbprint'],
            "authentication_scheme": {
                "scheme_name": "password",
                "username": data['username'],
                "password": data['password']
            }
        }
    }
    status, json, err, _ = rest_utils.local_node_request("POST", api, request)
    if status != compat23_httplib.OK:
        lg.error("Copy file error, api return: [status: %s, data: %s, err: %s]"
            % (status, json, err))
        return json['error_code'], "", json['error_message']
    return 0, "", ""


def _scp_file_with_progress(cmd, api=None, verbose=False):
    """ SCP file operation with progress status.
    Args:
        cmd:     (list) SCP file transfer command.
        api:     (str) specifies api path to use for scp tansfer.
        verbose: (bool) if enable, prints the output from scp to stdout.
    """
    import pexpect
    import getpass

    expectations = [os.path.basename(cmd[-2]), # basename of src file path
                    'password:',
                    'Password:',
                    'continue connecting (yes/no)?',
                    'Please type \'yes\' or \'no\': ',
                    pexpect.EOF,
                    pexpect.TIMEOUT]

    def _is_progress_status(line):
        progress_status = re.search(_PROGRESS_STATUS_PATTERN, line)
        if progress_status:
            return True
        False

    # add the flag to output the progress of operation
    if '-p' not in cmd:
        cmd.insert(1, '-p')
    cmd = " ".join(cmd)


    def _kill_child(child, retry=5):
        count = 0
        while count < retry:
            result = child.terminate(force=True)
            if result:
                return result
            count += 1
        return False


    def __copy_file_scp(child=None, thumbprint=None):
        if child is None:
            child = pexpect.spawn(cmd, timeout=_PEXPECT_TIMEOUT)
            initial = True
        else:
            initial = False
        res = child.expect(expectations)

        # In Python3, output returns byte strings
        line = child.before
        if line is not None:
            if is_py3():
                line = line.decode("utf8")

        if res == 0:
            # line with progress status, startswith src filename separated by details
            if _is_progress_status(line):
                sys.stdout.flush()
                sys.stdout.write(expectations[0] + line)
            return __copy_file_scp(child, thumbprint)
        elif res == 1:
            if not initial:
                line = "\n".join(line.split("\n")[1:])
            password = getpass.getpass(line + "password: ")
            if api is not None:
                scp_data = unpack_scp_cmd_values(cmd)
                # If cmd doesn't prompts to verify the ECDSA fingerprint
                # it must be already added into know_hosts
                if not thumbprint:
                    thumbprint = get_stored_thumbprint(scp_data['server'])
                scp_data['thumbprint'] = thumbprint
                scp_data['password'] = password

                # Terminate the scp child process and execute copy from api
                if not _kill_child(child):
                    lg.error("Unable to terminate child scp process")
                return _scp_file_api(scp_data, api)
            else:
                child.sendline(password)
            return __copy_file_scp(child, thumbprint)
        elif res == 2:
            if not initial:
                line = "\n".join(line.split("\n")[1:])
            password = getpass.getpass(line + "Password: ")
            if api is not None:
                scp_data = unpack_scp_cmd_values(cmd)
                if not thumbprint:
                    thumbprint = get_stored_thumbprint(scp_data['server'])
                scp_data['thumbprint'] = thumbprint
                scp_data['password'] = password
                if not _kill_child(child):
                    lg.error("Unable to terminate child scp process")
                return _scp_file_api(scp_data, api)
            else:
                child.sendline(password)
            return __copy_file_scp(child, thumbprint)
        elif res == 3:
            answer = get_compat23_input()(line + expectations[3] + " ")
            child.sendline(answer)
            if api is not None and thumbprint is None and answer == "yes":
                match = _FINGERPRINT_RE.match(line.replace("\n", ""))
                if match:
                    thumbprint = match.group(1)
                else:
                    log.error("No matching thumbprint pattern found \
                               in output: %s" % line)
            return __copy_file_scp(child, thumbprint)
        elif res == 4:
            child.sendline(get_compat23_input()(expectations[4]))
            return __copy_file_scp(child, thumbprint)
        elif res == 5:
            # End marker - SCP when completes the transfer successfully ends with 100%;
            # else incase of EOF marks an unsuccessful operation, with output string as error.
            if '100%' in line:
                sys.stdout.flush()
                sys.stdout.write(expectations[0] + line)
                err = ''
            else:
                if not initial:
                    line = "\n".join(line.split("\n"))
                if verbose:
                    sys.stdout.write(line)
                # Pexpect uses CR/LF combination for EOF
                err = line.replace('\r', '')
            child.close()
            return child.exitstatus, '', err
        child.close()
        return child.exitstatus, '', line
    return __copy_file_scp()


def _scp_file(src, dst, port=None, progress=True, api=None):
    """Invoke scp for specified source and destination files and optional port.

    Args:
        src: Source file to scp.
        dst: Destination file to scp.
        port: Optional port to use (instead of 22).
        api: specifies api path to use for scp tansfer.
             supported only on appliance type node.
                - /node/file-store?action=copy_to_remote_file
                - /node/core-dumps?action=copy_to_remote_file
    Return:
        (returncode, output, error) tuple where a non-zero returncode
        indicates an error, output and error are scp stdout and stderr
        output, respectively.
    Raises:
        KeyboardInterrupt
    """
    # Per #1414129 logged by iSec, we should not disable host identity
    # verification. Note that with host identity verification enabled,
    # scp/ssh may prompt the user for input, e.g. "Are you sure you want
    # to continue connecting (yes/no)?", which means that clients
    # invoking the copy command via pexpect or ssh will need to account
    # for the prompt. In the future we may want to extend the copy CLI
    # command with a ignore_unknown_host option, so commenting out the
    # insecure cmd invocation for now.
    #cmd = ["scp", "-o", "StrictHostKeyChecking no",
    #       "-o", "UserKnownHostsFile /dev/null"]
    # Also, set PreferredAuthentications to include both password and
    # keyboard-interactive to account for non-password authentication
    # that the remote server may prefer (e.g. if it sets
    # "PasswordAuthentication no" in its /etc/ssh/sshd_config file).
    cmd = ["scp", "-o",
           "PreferredAuthentications=password,keyboard-interactive"]
    port = port or 22   # set default scp port to 22
    cmd.extend(["-P", str(port)])
    cmd.extend([src, dst])
    rc, out, err = 1, "", ""
    node_type = get_node_type()
    # NSX appliances have default iptables DROP rules for outbound traffic on
    # unknown ports, so add an ALLOW rule here and remove it in finally block
    if is_appliance():
        if port and port != 22:
            ipt_cmd = ['sudo',
                       '/opt/vmware/nsx-cli/bin/scripts/ipt_rule_action.sh',
                       'set', '-p', str(port), '-t', 'TCP']
            po = Popen(ipt_cmd, stdout=PIPE, stderr=PIPE)
            out, err = po.communicate()
            rc = po.returncode
            lg.info("Creating iptable rule, "
                    "rc: %s, out: %s, err: %s", rc, out, err)
            if rc != 0:
                lg.error("Error creating iptable rule, "
                         "rc: %s, out: %s, err: %s", rc, out, err,
                         error_code=ec.OPERATION_FAILED)
                return (rc, out, err)

    # SCP with progress status do not supports on windows based platforms,
    # and requires pexpect to be available in environment.
    status_op = not is_windows() and progress and _is_pexpect_available()
    try:
        if status_op:
            rc, out, err = _scp_file_with_progress(cmd, api=api)
        else:
            lg.warn("SCP file operation with output progress not supported")
            po = Popen(cmd, stdout=PIPE, stderr=PIPE)
            out, err = po.communicate()
            if is_py3():
                # In Python3, communicate() returns byte strings,
                # so convert to string
                if out is not None:
                    out = out.decode("utf8")
                if err is not None:
                    err = err.decode("utf8")
            rc = po.returncode
        if rc:
            lg.error("Error in scp operation: return=%s, output=%s, error=%s"
                     % (rc, out, err))
    except KeyboardInterrupt:
        if not status_op:
            po.terminate()
            po.wait()  # Need to wait otherwise readline gets confused.
        lg.error("User interrupted file transfer operation: "
         "src=%s, dst=%s" % (os.path.basename(src), dst))
        raise
    except Exception:
        err = "Error invoking scp"
        lg.exception("%s: src=%s, dst=%s", err, src, dst,
                     error_code=ec.INTERNAL_ERROR)
    finally:
        if is_appliance():
            if port and port != 22:
                ipt_cmd = ['sudo',
                           '/opt/vmware/nsx-cli/bin/scripts/ipt_rule_action.sh',
                           'delete', '-p', str(port), '-t', 'TCP']
                po = Popen(ipt_cmd, stdout=PIPE, stderr=PIPE)
                out2, err2 = po.communicate()
                rc2 = po.returncode
                lg.info("Deleting iptable rule, "
                        "rc: %s, out: %s, err: %s", rc2, out2, err2)
        else:
            pass
    return (rc, out, err)


def _get_url_port(url):
    """Port property of url object returned from urlparse.

    Args:
        url: Url object returned from urlparse.
    Return:
        url.port value as implemented in Ubuntu 14.04's version of Python
        (Ubuntu 12.04's Python raises ValueError for the case where
        url is http://host:/path, Ubuntu 14.04's Python returns None).
    """
    try:
        return url.port
    except ValueError:
        return None


def get_filestore_files(sort_results=True):
    """Return list of filenames in filestore in optionally sorted order.

    Args:
        sort_results: True, if filenames should be sorted; otherwise, False.
    Return:
        List of filenames.
    """
    results = []
    status, obj, _, _ = local_node_request(
        "GET", "/node/file-store", retry_time=2)
    if status == compat23_httplib.OK:
        results = [item["name"] for item in obj.get("results", [])]
    else:
        # To support the command "copy file <filename> url <url>" to copy
        # techsupport bundles off box, we get the list of files directly
        # in case NAPI is unavailable. Ignore dotfiles - used for temporary
        # files during uploads.
        results = [fname for fname in os.listdir(_FILE_STORE_DIR)
                   if not fname.startswith(".")]
    if sort_results:
        results.sort()
    return results


class ExistingFileArgument(object):
    """Query argument to determine existing files in the filestore."""
    def __init__(self, descriptor):
        self._descriptor = descriptor

    def __call__(self, req_dict):
        """Retrieve existing files in the filestore.

        Return:
            List of existing files in the filestore.
        """
        return simplejson.dumps(get_filestore_files())


class HttpScpFileUrlArgument(ScpFileUrlArgument):
    _ALLOWED_SCHEMES = ["http", "scp", "https"]


class HostThumbprint(object):
    @classmethod
    def verify(cls, thumbprint, req_dict):
        return True


class GetFiles(CliCommandProvider):
    def __call__(self, cmd_list):
        results = []
        filter_filename = cmd_list[-1] if len(cmd_list) > 2 else None
        status, obj, _, _ = local_node_request("GET", "/node/file-store")
        if status != compat23_httplib.OK:
            return "MSG_FILESTORE_GET_FILES_FAILED"
        for item in obj.get("results", []):
            if filter_filename and filter_filename != item["name"]:
                continue
            results.append({
                "name": item["name"],
                "date": int(item["modified_epoch_ms"] / 1000),
                "permissions": "-rw-",  # CLI user has rw perms
                "size": item["size"]})
        if filter_filename and not results:
            return "MSG_FILESTORE_NOT_FOUND\n%s" % filter_filename
        return "OK\n\n%s" % simplejson.dumps(results)

    @staticmethod
    def sort_key(item):
        return item['name']


# DEPRECATED: delete after QE testcases use get file <file> thumbprint
class GetFileChecksum(CliCommandProvider):
    def __call__(self, cmd_list):
        filename = cmd_list[2]
        node_uri = "/node/file-store/%s/sha1" % filename
        status, obj, err, _ = local_node_request("GET", node_uri)
        if status == compat23_httplib.OK:
            return "OK\n\n%s" % simplejson.dumps(
                {"sha1sum": obj.get("hash_code")})
        elif status == compat23_httplib.NOT_FOUND:
            return "MSG_FILESTORE_NOT_FOUND\n%s" % filename
        return "MSG_FILESTORE_CHECKSUM_FAILED"


class GetFileThumbprint(CliCommandProvider):
    def __call__(self, cmd_list):
        filename = cmd_list[2]
        node_uri = "/node/file-store/%s/thumbprint" % filename
        status, obj, err, _ = local_node_request("GET", node_uri)
        if status == compat23_httplib.OK:
            return "OK\n\n%s" % simplejson.dumps(
                {"sha1sum": obj.get("sha1"),
                 "sha256sum": obj.get("sha256")})
        elif status == compat23_httplib.NOT_FOUND:
            return "MSG_FILESTORE_NOT_FOUND\n%s" % filename
        return "MSG_FILESTORE_CHECKSUM_FAILED"


class DelFile(CliCommandProvider):
    def __call__(self, cmd_list):
        pattern_or_filename = cmd_list[-1]
        status, obj, _, _ = local_node_request("GET", "/node/file-store")
        if status != compat23_httplib.OK:
            return "MSG_FILESTORE_GET_FILES_FAILED"

        data = obj.get("results", [])

        pattern = "^" + pattern_or_filename.replace("*", "[^/]*") + "$"
        target = [item['name'] for item in data
                  if re.match(pattern, item['name'])]

        if not target:
            return "MSG_FILESTORE_DELETE_NOT_FOUND\n%s" % pattern_or_filename

        failed = []
        for filename in target:
            if not delete_filestore_file(filename):
                failed.append(filename)
        return ("OK\n\n%s" % simplejson.dumps("") if not failed else
                "MSG_FILESTORE_DELETE_FAILED\n%s" % ", ".join(failed))


# NOTE: scp file copy requests are implemented by the CLI, all other copy
# requests invoke NAPI to perform the copy. scp is implemented locally
# because if the authenticity of the remote host has not been verified
# (i.e. the remote host fingerprint is not already in the .ssh/known_hosts
# file or conflicts with what's there), we want to prompt the user to
# accept the remote host.
class CopyFileFrom(CliCommandProvider):
    def __call__(self, cmd_list):
        filestore = get_file_store()
        specified_url = cmd_list[2]
        url = compat23_urlparse(specified_url)
        if url.scheme == "https":
            #https command specified with filename
            if len(cmd_list) > 6:
                filename = cmd_list[-1]
            else:
                #https command specified without file name
                filename = os.path.basename(cmd_list[2])
        elif len(cmd_list) > 3:
            # Command specified with "file <filename>"
            filename = cmd_list[-1]
        else:
            # Command specified without "file <filename>" use the basename
            # of the specified <url> for the filename
            filename = os.path.basename(cmd_list[-1])
        dst_path = os.path.abspath(os.path.join(filestore, filename))
        if os.path.isdir(dst_path):
            return "MSG_FILESTORE_DIRECTORY_SPECIFIED\n%s" % url.path
        elif os.path.isfile(dst_path):
            return "MSG_FILESTORE_ALREADY_EXISTS\n%s" % filename
        elif not dst_path.startswith(filestore + "/"):
            # The <filename-arg> that receives the user specified filename
            # should already check that the filename doesn't contain a
            # relative path, but we check again just in case.
            return "MSG_FILESTORE_FILE_NOT_ALLOWED\n%s" % filename
        rc = None
        specified_url = cmd_list[2]
        url = compat23_urlparse(specified_url)
        port = _get_url_port(url)
        if url.scheme == "scp":
            src_path = "%s@%s:%s" % (url.username, url.hostname, url.path)
            tmp_fd, tmp_path = tempfile.mkstemp(dir=get_tmp_store())
            os.close(tmp_fd)
            try:
                rc, out, err = _scp_file(src_path, tmp_path, port)
                if rc == 0:
                    result, err_msg_id = move_file_to_filestore(
                        os.path.basename(tmp_path), filename)
                    if result:
                        return "OK\n\n%s" % simplejson.dumps("")
                    elif err_msg_id:
                        return err_msg_id
                elif (isinstance(err, compat23_string_types) and
                        _SIG_NO_SPACE in err):
                    return "MSG_COPY_NO_SPACE_ERROR"
                elif (isinstance(err, compat23_string_types) and
                        _PERMISSION_DENIED in err):
                    return "MSG_COPY_PERMISSION_DENIED"
                elif (isinstance(err, compat23_string_types) and
                        _AUTHENTICATION_FAIL in err):
                    return "MSG_COPY_PERMISSION_DENIED"
                elif (isinstance(err, compat23_string_types) and
                        _INVALID_PATH in err):
                    return "MSG_FILESTORE_INVALID_PATH\n%s" % url.path
                else:
                    lg.error(
                        "Error copying remote file %s to filestore, rc=%s, "
                        "out=%s, err=%s" % (src_path, rc, out, err),
                        error_code=ec.OPERATION_FAILED)
            finally:
                delete_file_if_exists(tmp_path)
        elif url.scheme == "http":
            req_obj = {
                "protocol": {"name": "http"},
                "server": url.hostname,
                "uri": url.path}
            if port:
                req_obj["port"] = port
            if url.username:
                req_obj["authentication_scheme"] = {
                    "scheme_name": "basic",
                    "username": url.username,
                    "password": url.password or ""}
            uri = "/node/file-store/%s?action=copy_from_remote_file" % filename
            status, obj, err, _ = local_node_request("POST", uri, req_obj)
            if status == compat23_httplib.CREATED:
                return "OK\n\n%s" % simplejson.dumps("")
            elif status == compat23_httplib.BAD_REQUEST:
                return "MSG_FILESTORE_INVALID_SERVER"
            elif status == compat23_httplib.NOT_FOUND:
                return "MSG_FILESTORE_NOT_FOUND\n%s" % specified_url
            elif status == compat23_httplib.UNAUTHORIZED:
                return "MSG_FILESTORE_UNAUTHORIZED"
            elif (status == compat23_httplib.SERVICE_UNAVAILABLE and
                    obj.get("error_code") == 36002):
                return "MSG_COPY_NO_SPACE_ERROR"
            else:
                lg.error(
                    "Error copying http file, status: %s, obj: %s, err: %s",
                    status, obj, err, error_code=ec.OPERATION_FAILED)
        elif url.scheme == "https":
            if port:
                host_port = port
            else:
                host_port = 443
            if len(cmd_list) < 5:
                lg.error("Insufficient number of arguments provided.")
                return "MSG_HTTPS_FORMAT_ERROR"
            host_thumbprint = cmd_list[4]
            if not host_thumbprint:
                return "MSG_FILESTORE_COPY_ERROR\n%s" % "Couldn't fetch thumbprint"
            req_obj = {
                "protocol": {"name": "https",
                             "sha256_thumbprint": host_thumbprint},
                "server": url.hostname,
                "uri": url.path}
            if port:
                req_obj["port"] = port
            if url.username:
                req_obj["authentication_scheme"] = {
                    "scheme_name": "basic",
                    "username": url.username,
                    "password": url.password or ""}
            uri = "/node/file-store/%s?action=copy_from_remote_file" % filename
            status, obj, err, _ = local_node_request("POST", uri, req_obj)
            if status == compat23_httplib.CREATED:
                return "OK\n\n%s" % simplejson.dumps("")
            elif status == compat23_httplib.BAD_REQUEST:
                return "MSG_FILESTORE_INVALID_SERVER"
            elif status == compat23_httplib.NOT_FOUND:
                return "MSG_FILESTORE_NOT_FOUND\n%s" % specified_url
            elif status == compat23_httplib.UNAUTHORIZED:
                return "MSG_FILESTORE_UNAUTHORIZED"
            elif (status == compat23_httplib.SERVICE_UNAVAILABLE and
                    obj.get("error_code") == 36002):
                return "MSG_FILESTORE_NO_SPACE_ERROR"
            else:
                lg.error(
                    "Error copying https file, status: %s, obj: %s, err: %s",
                    status, obj, err, error_code=ec.OPERATION_FAILED)
        else:
            err = "Unexpected copy file url scheme: %s" % url.scheme
            lg.error(err, error_code=ec.INVALID_ARGUMENT)

        return "MSG_FILESTORE_COPY_ERROR\n%s\n%s\n%s" % (
                    specified_url, filename, ' '.join(err.splitlines()))


class CopyFileTo(CliCommandProvider):
    def __call__(self, cmd_list):
        filename = cmd_list[2]
        src_path = os.path.join(get_file_store(), filename)
        if os.path.isdir(src_path):
            return "MSG_FILESTORE_DIRECTORY_SPECIFIED\n%s" % filename
        elif not os.path.isfile(src_path):
            return "MSG_FILESTORE_NOT_FOUND\n%s" % filename
        rc = None
        specified_url = cmd_list[-1]
        url = compat23_urlparse(specified_url)
        if url.scheme == "scp":
            dst_path = "%s@%s:%s" % (url.username, url.hostname, url.path)
            rc, out, err = _scp_file(src_path, dst_path, _get_url_port(url))
        else:
            # Note: we currently only support copying a local file to a
            # remote destination via scp.
            err = "Unexpected copy file url scheme: %s" % url.scheme
            lg.error(err, error_code=ec.INVALID_ARGUMENT)
        if rc == 0 and err == "":
            return "OK\n\n%s" % simplejson.dumps("")
        elif rc == 0 and err.startswith("Warning"):
            lg.warning("Copying file %s to %s: %s" % (
                filename, dst_path, ' '.join(err.splitlines())))
            return "OK\n\n%s" % simplejson.dumps("")
        elif _SIG_HOST_KEY_ERROR in err:
            return "MSG_COPY_HOST_KEY_ERROR"
        elif _SIG_NO_SPACE in err:
            return "MSG_COPY_NO_SPACE_ERROR"
        elif (isinstance(err, compat23_string_types) and
                _PERMISSION_DENIED in err):
            return "MSG_COPY_PERMISSION_DENIED"
        elif (isinstance(err, compat23_string_types) and
                _AUTHENTICATION_FAIL in err):
            return "MSG_COPY_PERMISSION_DENIED"

        return "MSG_FILESTORE_COPY_ERROR\n%s\n%s\n%s" % (
                    filename, dst_path, ' '.join(err.splitlines()))
