# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import argparse
import json
import socket
import sys
import uuid

from sha.cli import socket_utils
from sha.cli.cli_utils import CLI_REQ_ID, CLI_REQ_CMD, CLI_REQ_ARGS, \
    CLI_RESP_RESULT, CLI_RESP_ERR_CODE, CLI_RESP_ERR_DESC, \
    CLI_SERVICE_SOCKET, CLI_CONTROL_SOCKET, CLI_REQ_HELP
from sha.cli.cli_cmds import CLI_CMD_LIST_CMD


def parse_additional_args(extend_args):
    if len(extend_args) % 2 != 0:
        print("Key/value of additional argument is not consistent", file=sys.stderr)
        sys.exit(1)
    index = 0
    args = {}
    while index < len(extend_args):
        key = extend_args[index]
        value = extend_args[index + 1]
        if not key.startswith('--'):
            print("Key of additional argument should start with '--'", file=sys.stderr)
            sys.exit(1)
        key = key[2:]
        args[key] = value
        index += 2
    return args


def request(req):
    conn = None
    try:
        conn = socket_utils.connect_socket(CLI_SERVICE_SOCKET)
        status, err_msg = socket_utils.send_msg(conn, json.dumps(req))
        if not status:
            return False, err_msg
        msg, err_msg = socket_utils.receive_msg(conn)
        if err_msg:
            return False, err_msg
        return True, json.loads(msg)
    except socket.error as e:
        return False, e.strerror
    finally:
        socket_utils.close_conn(conn)


def notify(req_id):
    conn = None
    try:
        conn = socket_utils.connect_socket(CLI_CONTROL_SOCKET)
        req = {
            CLI_REQ_ID: req_id
        }
        status, err_msg = socket_utils.send_msg(conn, json.dumps(req))
        if not status:
            return False, err_msg
        else:
            return True, None
    except Exception as e:
        return False, str(e)
    finally:
        socket_utils.close_conn(conn)


def main():
    parser = argparse.ArgumentParser(prog="sha-appctl",
                                     epilog="""
                                     Append '-?' after a command line argument to get the help of certain command.
                                     """)
    parser.add_argument('-c', '--cmd', type=str, help='command name')
    parser.add_argument('-l', '--list', action='store_true', help='list supported commands')
    basic_args, extend_args = parser.parse_known_args()
    req = {
        CLI_REQ_ID: str(uuid.uuid4()),
        CLI_REQ_CMD: basic_args.cmd
    }
    if basic_args.list:
        req[CLI_REQ_CMD] = CLI_CMD_LIST_CMD
    else:
        if not basic_args.cmd:
            print("Argument -c/--cmd is missing", file=sys.stderr)
            sys.exit(1)
        req[CLI_REQ_CMD] = basic_args.cmd
        if '-?' in extend_args:
            parsed_args = {
                CLI_REQ_HELP: True
            }
        else:
            parsed_args = parse_additional_args(extend_args)
        if parsed_args:
            req[CLI_REQ_ARGS] = parsed_args
    try:
        status, result = request(req)
    except KeyboardInterrupt as e:
        _, err_msg = notify(req[CLI_REQ_ID])
        if err_msg:
            print(err_msg, file=sys.stderr)
        raise e
    if status:
        if int(result[CLI_RESP_ERR_CODE]) == 0:
            result = result[CLI_RESP_RESULT]
            print(json.dumps(result), file=sys.stdout)
        else:
            result = result[CLI_RESP_ERR_DESC]
            print(result, file=sys.stderr)
    else:
        print(result, file=sys.stderr)


if __name__ == "__main__":
    main()
