# Copyright (C) 2020-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
from multiprocessing import Process, Pipe
from threading import Thread

import errno
import os
import signal
import subprocess
import sys
import time

from concurrent.futures import ThreadPoolExecutor
from queue import Queue, Empty

import sha.preconfig
from sha.constants import FORKSERVER_MONITOR_STATS, FORKSERVER_STATS, FORKSERVER_PROCESS_STATS, FORKSERVER_EXECUTOR_STATS

# CMD Request/Response Format
# Request:
# {
#    'seq': sequence number,
#    'type': command type
#    'cmd': cmd_string,
#    'input': interactive input, must be binary sequence
#    'shell': boolean,
#    'timeout': timeout seconds,
#    'resp_queue": queue to transfer the resp from forkserver
#    'check': boolean, keep-alive check req
#    'timestamp': timestamp of command requested or submitted to forkserver
# }
#
# Response:
# {
#    'seq': sequence number,
#    'type': command type
#    'output': output returned by subprocess,
#    'exception': catched exception,
#    'error': error generated by forkserver process
#    'excutor': executor index
#    'timestamp': timestamp of commands request reach executor
#    'execute_time': execute time
#    'stats': forkserver stats info
# }
CMD_SEQUENCE = 'seq'
CMD_TYPE = 'type'
CMD_REQUEST_COMMAND = 'cmd'
CMD_REQUEST_INPUT = 'input'
CMD_REQUEST_SHELL = 'shell'
CMD_REQUEST_TIMEOUT = 'timeout'
CMD_REQUEST_TIMED_OUT = 'timed_out'
CMD_REQUEST_RESP_QUEUE = 'resp_queue'
CMD_REQUEST_CHECK = 'check'
CMD_REQUEST_TIMESTAMP = 'timestamp'
CMD_REQUEST_ENV = 'env'
CMD_RESPONSE_OUTPUT = 'output'
CMD_RESPONSE_EXCEPTION = 'exception'
CMD_RESPONSE_ERROR = 'error'
CMD_RESPONSE_EXECUTOR = 'executor'
CMD_RESPONSE_EXECUTE_TIME= 'execute_time'
CMD_RESPONSE_STATS = 'stats'

# Allow at max 5 parallel commands executed
FORK_SERVER_EXCUTOR_MAX_COUNT = 5

# Command request type
# commands from pre-defined plugin/metric or other threads,
# and execute time not bigger than FORK_SERVER_CMD_EXECUTE_DEFAULT_MAX_SEC
CMD_TYPE_DEFAULT = 0
# commands from custom plguin or need more than FORK_SERVER_CMD_EXECUTE_DEFAULT_MAX_SEC to execute
CMD_TYPE_HEAVY = 1
CMD_TYPE_MAX = 2

# Considering commands in executors, pipe and req/resp queue of forksever
# process, allow FORK_SERVER_EXCUTOR_MAX_COUNT * 4 default(pre-defined) command requests
# sent to forkserver process at the same time.
DEFAULT_REQUESTS_MAX_COUNT = FORK_SERVER_EXCUTOR_MAX_COUNT * 4
# For other(custom) type requests, only allow at max FORK_SERVER_EXCUTOR_MAX_COUNT
OTHER_REQUESTS_MAX_COUNT = FORK_SERVER_EXCUTOR_MAX_COUNT

requests_max_count=[DEFAULT_REQUESTS_MAX_COUNT, OTHER_REQUESTS_MAX_COUNT]

# In order to support all metric collectors/dynamic plugins running
# in a parallel way, the queue size should be enough to contain
# one request from each of them
FOEK_SERVER_REQUEST_QUEUE_SIZE = 100

# Forkserver loop check interval in second
FORK_SERVER_BUSY_CHECK_INTERVAL = 0.01
FORK_SERVER_IDLE_CHECK_INTERVAL = 1

# Max seconds of requests pending before sent to forkserver process
FORK_SERVER_CMD_PENDING_SEC = 2
# Default command execute time, follow the value in metric
FORK_SERVER_CMD_EXECUTE_DEF_SEC = 4
# Max allowed execute time for DEFAULT commands
# Take commands need more time as "HEAVY" commands
FORK_SERVER_CMD_EXECUTE_DEFAULT_MAX_SEC = 10
# max allowed execute time for heavy commands
FORK_SERVER_CMD_EXECUTE_HEAVY_MAX_SEC = 60
# Delta seconds introduced by forkserver process
FORK_SERVER_CMD_DELTA_SEC = 2

# forkserver keep-alive heartbeat interval in second
FORK_SERVER_KEEP_ALIVE_DEFAULT_INTERVAL = 20  # Default command
FORK_SERVER_KEEP_ALIVE_HEAVY_INTERVAL = 60    # Heavy command
FORK_SERVER_KEEP_ALIVE_MAX_RETRY = 6

keep_alive_interval = [FORK_SERVER_KEEP_ALIVE_DEFAULT_INTERVAL, FORK_SERVER_KEEP_ALIVE_HEAVY_INTERVAL]

FORK_SERVER_ERROR_INVALID_REQUEST = 'Invalid request'
FORK_SERVER_ERROR_REQ_QUEUE_FULL = 'No resource in fork server: req queue full'
FORK_SERVER_ERROR_PENDING_TIMEOUT = 'Request timeout when waiting for resource'
FORK_SERVER_ERROR_EXECUTE_TIMEOUT = 'Request timeout when waiting for response'
FORK_SERVER_ERROR_SENT_FAILURE = 'No resource in fork server: failed to send req'

# statistics info
REQ = 'req'
REQ_SENT = 'req_sent'
REQ_ERROR = 'req_error'
REQ_DROP_NO_RESOURCE = 'req_dropped_no_resource'
REQ_DROP_TIMEOUT = 'req_dropped_timeout'
RESP = 'resp'
RESP_RCVD = 'resp_rcvd'
RESP_ERROR = 'resp_error'
RESP_DROPPED_NO_REQ = 'resp_dropped_no_req'
KEEP_ALIVE_REQ = 'keep_alive_req'
KEEP_ALIVE_RESP = 'keep_alive_resp'
MAPPING_REQ_PEAK = 'peak_mapping_req'
PENDING_REQ_PEAK = 'peak_pending_req'
PENDING_REQ = 'pending_req'
SEND_EAGAIN = 'send_eagain'
RECV_EAGAIN = 'recv_eagain'

# Executor param dict
EXECUTOR_REQ_QUEUE = 'req_queue'
EXECUTOR_RESP_QUEUE = 'resp_queue'
EXECUTOR_EXIT_FLAG = 'exit'

# EAGAIN/"[Errno 11] Resource temporarily unavailable" process
EAGAIN_MAX_COUNT = 6
EAGAIN_WAIT_INTERVAL = 1

eagain_recv_count = 0
eagain_send_count = 0

def fork_server_sleep(watcher, seconds):
    sleep_seconds = seconds
    # check the exit_now flag to exit in time
    while not watcher.exit_now and sleep_seconds > 0:
        time.sleep(0.1)
        sleep_seconds -= 0.1

def conn_recv(conn, stats, logger, watcher):
    global eagain_recv_count
    try:
        msg = conn.recv()
        # reset eagain_count
        eagain_recv_count = 0
        return msg
    except BlockingIOError as e:
        if e.errno != errno.EAGAIN:
            # reset if not failed by [Errno 11]/EAGAIN, raise as before
            eagain_count = 0
            raise e
        stats[RECV_EAGAIN] += 1
        eagain_recv_count += 1
        logger.warning('Encountering {} the {}th time '
                       'when receiving message'.format(e, eagain_recv_count))
        if (eagain_recv_count >= EAGAIN_MAX_COUNT):
            # if [Errno 11]/EAGAIN happened continuously EAGAIN_MAX_COUNT, raise the exception
            raise e
        # [Errno 11] happened less than EAGAIN_MAX_COUNT, ignore it and sleep EAGAIN_WAIT_INTERVAL
        fork_server_sleep(watcher, EAGAIN_WAIT_INTERVAL)
        return None

def conn_send(conn, msg, stats, logger, watcher):
    global eagain_send_count
    try:
        conn.send(msg)
        eagain_send_count = 0
        return True
    except BlockingIOError as e:
        if e.errno != errno.EAGAIN:
            # reset if not failed by [Errno 11]/EAGAIN, raise as before
            eagain_send_count = 0
            raise e
        stats[SEND_EAGAIN] += 1
        eagain_send_count += 1
        logger.warning('Encountering {} the {}th time '
                       'when sending message: {}'.format(e, eagain_send_count, msg))
        if (eagain_send_count >= EAGAIN_MAX_COUNT):
            # if [Errno 11]/EAGAIN happened continuously EAGAIN_MAX_COUNT, raise the exception
            raise e
        # [Errno 11] happened less than EAGAIN_MAX_COUNT, ignore it and sleep EAGAIN_WAIT_INTERVAL
        fork_server_sleep(watcher, EAGAIN_WAIT_INTERVAL)
        return False

def conn_send_with_retry(conn, msg, stats, logger, watcher):
    if not conn_send(conn, msg, stats, logger, watcher):
        logger.info('Retry to send msg :{}'.format(msg))
        return conn_send(conn, msg, stats, logger, watcher)
    return True

class SignalWatcher:
    def __init__(self):
        self.exit_now = False
        signal.signal(signal.SIGINT, self._set_kill)
        signal.signal(signal.SIGTERM, self._set_kill)
        signal.signal(signal.SIGABRT, self._set_kill)

    def _set_kill(self, signum, frame):
        self.exit_now = True

def _executor_get_req(req_queues):
    i = 0
    for queue in req_queues:
        try:
            if i == 0:
                 return queue.get(timeout=0.1)
            else:
                 return queue.get_nowait()
        except Empty:
            i += 1

    raise Empty

def _fork_server_executor(index, executor_param, req_queues, stats, logger):
    """ _forkserver executor
    """
    resp_queue = executor_param[EXECUTOR_RESP_QUEUE]
    while not executor_param[EXECUTOR_EXIT_FLAG]:
        try:
            cmd_req = _executor_get_req(req_queues)
            stats[REQ] += 1
        except Empty:
            continue
        logger.debug('get cmd_req from queue: {}'.format(cmd_req))
        cmd_resp = {
            CMD_SEQUENCE: cmd_req[CMD_SEQUENCE],
            CMD_TYPE: cmd_req[CMD_TYPE],
            CMD_RESPONSE_EXECUTOR: index,
            CMD_REQUEST_TIMESTAMP: time.monotonic()
        }
        # keep-alive check request, return resp directly
        if cmd_req.get(CMD_REQUEST_CHECK, False):
            cmd_resp[CMD_RESPONSE_STATS] = executor_param[CMD_RESPONSE_STATS]
            resp_queue.put(cmd_resp)
            stats[RESP] += 1
            logger.debug('put cmd_resp into queue: {}'
                         .format(cmd_resp))
            continue

        input = cmd_req.get(CMD_REQUEST_INPUT, None)

        # execute command
        try:
            if not input:
                output = subprocess.check_output(
                         cmd_req[CMD_REQUEST_COMMAND],
                         stderr=subprocess.PIPE,
                         shell=cmd_req[CMD_REQUEST_SHELL],
                         timeout=cmd_req[CMD_REQUEST_TIMEOUT],
                         env=cmd_req[CMD_REQUEST_ENV]
                     )
            else:
                output = subprocess.run(
                         cmd_req[CMD_REQUEST_COMMAND],
                         input=input,
                         check=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         shell=cmd_req[CMD_REQUEST_SHELL],
                         timeout=cmd_req[CMD_REQUEST_TIMEOUT],
                         env=cmd_req[CMD_REQUEST_ENV]
                     ).stdout

            cmd_resp[CMD_RESPONSE_EXECUTE_TIME] = time.monotonic() - cmd_resp[CMD_REQUEST_TIMESTAMP]
            cmd_resp[CMD_RESPONSE_OUTPUT] = output
        except Exception as e:
            cmd_resp[CMD_RESPONSE_EXCEPTION] = e
            logger.info("Exception catched when running cmd '{}': {}"
                        .format(cmd_req, cmd_resp))

        resp_queue.put(cmd_resp)
        stats[RESP] += 1
        logger.debug('put cmd_resp into queue: {}'.format(cmd_resp))

    logger.warning('forkserver executor[{}] stopped'.format(index))


def fork_server_executor(index, executor_param, req_queues, stats):
    """ forkserver executor
    """
    logger = sha.preconfig.get_sub_logger(__name__, 'fork-executor-{}'.format(index))
    logger.info('forkserver executor {} started: {}, {}'.format(index, executor_param, req_queues))
    try:
        _fork_server_executor(index, executor_param, req_queues, stats, logger)
    except Exception as e:
        import traceback
        trace = traceback.format_exc()
        logger.info("failed: {}, {}".format(e, trace))


def _fork_server_loop_impl(conn, req_queues, resp_queue, stats, logger):
    """ Forkserver loop implementation
    """
    watcher = SignalWatcher()
    keep_alive_timestamp = time.monotonic()

    pending_req_count = 0
    while not watcher.exit_now:
        # Get next request
        while conn.poll():
            cmd_req = conn_recv(conn, stats, logger, watcher)
            if not cmd_req:
                # skip to next loop to retry
                continue
            logger.debug('cmd_req recved: {}'.format(cmd_req))
            cmd = cmd_req.get(CMD_REQUEST_COMMAND, None)
            check = cmd_req.get(CMD_REQUEST_CHECK, False)
            seq = cmd_req.get(CMD_SEQUENCE, None)
            cmd_type = cmd_req[CMD_TYPE]
            # Update keep-alive heartbeat timestamp
            if check:
                keep_alive_timestamp = time.monotonic()
            if seq is not None and (cmd is not None or check):
                # The queue shouldn't be full in normal status.
                if not req_queues[cmd_type].full():
                    req_queues[cmd_type].put(cmd_req)
                    stats[REQ] += 1
                    pending_req_count += 1
                    stats[PENDING_REQ] = pending_req_count
                    if pending_req_count > stats[PENDING_REQ_PEAK]:
                        stats[PENDING_REQ_PEAK] = pending_req_count
                        logger.info("pending_req_peak changed: {}".format(pending_req_count))
                    logger.debug('put cmd_req into queue: {}'.format(cmd_req))
                else:
                    logger.warning('No resource in forkserver: {}'.format(cmd_req))
                    stats[REQ_DROP_NO_RESOURCE] += 1
                    cmd_resp = {CMD_SEQUENCE: seq, CMD_TYPE: cmd_type}
                    cmd_resp[CMD_RESPONSE_ERROR] = FORK_SERVER_ERROR_REQ_QUEUE_FULL
                    if conn_send_with_retry(conn, cmd_resp, stats, logger, watcher):
                        stats[RESP] += 1
            else:
                logger.warning('Invalid request: {}'.format(cmd_req))
                stats[REQ_ERROR] += 1
                cmd_resp = {CMD_SEQUENCE: seq, CMD_TYPE: cmd_type}
                cmd_resp[CMD_RESPONSE_ERROR] = FORK_SERVER_ERROR_INVALID_REQUEST
                if conn_send_with_retry(conn, cmd_resp, stats, logger, watcher):
                    stats[RESP] += 1
                    logger.debug('cmd_resp sent: {}'.format(cmd_resp))

        while not resp_queue.empty():
            cmd_resp = resp_queue.get()
            pending_req_count -= 1
            stats[PENDING_REQ] = pending_req_count
            logger.debug('get cmd_resp from queue: {}'.format(cmd_resp))
            if conn_send_with_retry(conn, cmd_resp, stats, logger, watcher):
                stats[RESP] += 1
                logger.debug('cmd_resp sent: {}'.format(cmd_resp))

        # Check keep-alive heartbeat
        if (time.monotonic() - keep_alive_timestamp > \
            (FORK_SERVER_KEEP_ALIVE_DEFAULT_INTERVAL * FORK_SERVER_KEEP_ALIVE_MAX_RETRY +
             FORK_SERVER_CMD_DELTA_SEC)):
            logger.warning("No keep-alive heartbeat for {} seconds. Exit."
                           .format(time.monotonic() - keep_alive_timestamp))
            break

        # Wait for new req comming
        try:
            conn.poll(timeout=(FORK_SERVER_BUSY_CHECK_INTERVAL if pending_req_count else \
                               FORK_SERVER_IDLE_CHECK_INTERVAL))
        except OverflowError as e:
            import traceback
            trace = traceback.format_exc()
            logger.info("conn poll OverflowError, maybe due to system clock change: {}, {}".format(e, trace))
            if sha.preconfig.SYSTEM != sha.preconfig.PLATFORM_WINDOWS:
                # After system clock changed, SHA main process may hanging when exit gracefully, kill it.
                logger.error("kill the SHA main process forcely:{}".format(os.getppid()))
                os.kill(os.getppid(), signal.SIGKILL)
            raise e


def fork_server(conn):
    """ Forkserver main thread
    """
    logger = sha.preconfig.get_sub_logger(__name__, 'fork-server')
    logger.info('process of forkserver starting')
    # create executor threads
    executor = ThreadPoolExecutor(max_workers=FORK_SERVER_EXCUTOR_MAX_COUNT)
    req_queues = []
    for cmd_type in range(CMD_TYPE_MAX):
        # Allow one more pre-defined req for keep-alive check
        req_queue = Queue(requests_max_count[cmd_type] + 1)
        req_queues.append(req_queue)
    resp_queue = Queue(FORK_SERVER_EXCUTOR_MAX_COUNT)
    executor_param = {
        EXECUTOR_RESP_QUEUE: resp_queue,
        EXECUTOR_EXIT_FLAG: False
    }
    stats = {
       REQ: 0,
       RESP: 0,
    }
    future_list = []
    executor_stats = {}
    for index in range(FORK_SERVER_EXCUTOR_MAX_COUNT):
       new_stats = stats.copy()
       # executor 0, 1, 2 dedicated for CMD_TYPE_DEFAULT
       # executor 3, 4 process CMD_TYPE_HEAVY in priority. Also handle
       # CMD_TYPE_DEFAULT when free.
       if index < 3:
          req_queue_list =[req_queues[CMD_TYPE_DEFAULT]]
       else:
          req_queue_list =[req_queues[CMD_TYPE_HEAVY], req_queues[CMD_TYPE_DEFAULT]]
       future = executor.submit(fork_server_executor, index, executor_param, req_queue_list, new_stats)
       future_list.append(future)
       executor_stats[index] = new_stats

    logger.info('process of forkserver started')
    logger.info('future list {}'.format(future_list))

    stats = {
        REQ: 0,
        RESP: 0,
        PENDING_REQ: 0,
        PENDING_REQ_PEAK: 0,
        REQ_ERROR: 0,
        RESP_ERROR: 0,
        REQ_DROP_NO_RESOURCE: 0,
        RECV_EAGAIN : 0,
        SEND_EAGAIN : 0,
    }
    executor_param[CMD_RESPONSE_STATS] = [stats, executor_stats]
    try:
        _fork_server_loop_impl(conn, req_queues, resp_queue, stats, logger)
    except Exception as e:
        import traceback
        trace = traceback.format_exc()
        logger.info("failed: {}, {}".format(e, trace))

    # Graceful exit process
    executor_param[EXECUTOR_EXIT_FLAG] = True

    for future in future_list:
       future.cancel()

    retry = 5
    while retry:
        # empty resp queue to make space for executors
        while not resp_queue.empty():
            resp_queue.get()
        # Check executors' state
        for i in range(len(future_list) - 1, -1, -1):
            if future_list[i].done():
                logger.warning('future {} of forkserver stopped'.format(future_list[i]))
                del future_list[i]
        # Skip the loop if all executors stopped
        if not future_list:
            break

        time.sleep(0.1)
        retry -= 1

    conn.close()

    logger.warning('process of forkserver stopped')
    logger.info('Fork server process stats: {}, executor stats: {} '.format(stats, executor_stats))


class ForkServerMonitor(Thread):
    """ Monitor thread running on the main process
    """

    def __init__(self, server_process, server_conn, req_queue, *args, **kwargs):
        super().__init__(name='nsx-sha-forkserver-monitor', *args, **kwargs)
        self._logger = sha.preconfig.get_sub_logger(__name__, 'fork-monitor')
        self._logger.info("forkserver monitor initializing")
        self._exit_now = False
        self._process = server_process
        self._conn = server_conn
        self._req_queue = req_queue
        self._mapping_dicts = []
        self._pending_lists = []
        for cmd_type in range(CMD_TYPE_MAX):
           self._mapping_dicts.append({})
           self._pending_lists.append([])
        self._timeout_check_timestamp = time.monotonic()
        self._keep_alive_timestamp = [time.monotonic(), time.monotonic()]
        self._keep_alive_retry =[FORK_SERVER_KEEP_ALIVE_MAX_RETRY, FORK_SERVER_KEEP_ALIVE_MAX_RETRY]
        self._cmd_seq = 0
        self._stats = {
            REQ: 0,
            KEEP_ALIVE_REQ: 0,
            REQ_SENT: 0,
            REQ_DROP_NO_RESOURCE: 0,
            REQ_DROP_TIMEOUT: 0,
            RESP: 0,
            KEEP_ALIVE_RESP: 0,
            RESP_RCVD: 0,
            REQ_ERROR: 0,
            RESP_ERROR: 0,
            RESP_DROPPED_NO_REQ: 0,
            MAPPING_REQ_PEAK: [0, 0],
            PENDING_REQ_PEAK: [0, 0],
            RECV_EAGAIN : 0,
            SEND_EAGAIN : 0,
        }


    def _has_resource(self, req_type):
        return  len(self._mapping_dicts[req_type]) < requests_max_count[req_type]

    def _pend_req(self, cmd_req, timestamp):
        cmd_req[CMD_REQUEST_TIMESTAMP] = timestamp
        cmd_type = cmd_req[CMD_TYPE]
        self._pending_lists[cmd_type].append(cmd_req)
        if len(self._pending_lists[cmd_type]) > self._stats[PENDING_REQ_PEAK][cmd_type]:
            self._stats[PENDING_REQ_PEAK][cmd_type] = len(self._pending_lists[cmd_type])
            self._logger.info('peak_pending_req[{}] changed: {}'
                              .format(cmd_type, self._stats[PENDING_REQ_PEAK][cmd_type]))

    def _send(self, cmd_req, map_req):
        is_sent = conn_send(self._conn, cmd_req, self._stats, self._logger, self)
        if is_sent:
            # Put the req into mapping dict to wait for response
            self._stats[REQ_SENT] += 1
            self._logger.debug('cmd_req sent: {}'.format(cmd_req))
            # Put the map_req into the mapping dict
            cmd_type = cmd_req[CMD_TYPE]
            self._mapping_dicts[cmd_type][self._cmd_seq] = map_req
            if len(self._mapping_dicts[cmd_type]) > self._stats[MAPPING_REQ_PEAK][cmd_type]:
                self._stats[MAPPING_REQ_PEAK][cmd_type] = len(self._mapping_dicts[cmd_type])
                self._logger.info('peak_mapping_req[{}] changed: {}'
                                  .format(cmd_type, self._stats[MAPPING_REQ_PEAK][cmd_type]))
        # Update seqence
        if self._cmd_seq == sys.maxsize:
            self._cmd_seq = 0
        else:
            self._cmd_seq += 1
        return is_sent

    def _send_req(self, cmd_req, timestamp):
        # The original request contains the resp queue, which cannot been sent via Pipe
        # Need create a new request and send it.
        cmd_req[CMD_REQUEST_TIMESTAMP] = timestamp
        cmd_req[CMD_SEQUENCE] = self._cmd_seq
        new_req = cmd_req.copy()
        del new_req[CMD_REQUEST_RESP_QUEUE]
        return self._send(new_req, cmd_req)

    def _report_stats(self, stats_list=None):
        from sha.core.whiteboard import DataWhiteboard
        self._logger.debug('forkserver monitor stats: {}'.format(self._stats))
        if not DataWhiteboard.report_data(FORKSERVER_MONITOR_STATS, self._stats):
            self._logger.info('failed to report {} to whiteboard'.format(FORKSERVER_MONITOR_STATS))
        self._logger.debug('ForkServer stats: {}'.format(ForkServer._stats))
        if not DataWhiteboard.report_data(FORKSERVER_STATS, ForkServer._stats):
            self._logger.info('failed to report {} to whiteboard'.format(FORKSERVER_STATS))
        if stats_list:
            if not DataWhiteboard.report_data(FORKSERVER_PROCESS_STATS, stats_list[0]):
                self._logger.info('failed to report {} to whiteboard'.format(FORKSERVER_PROCESS_STATS))
            if not DataWhiteboard.report_data(FORKSERVER_EXECUTOR_STATS, stats_list[1]):
                self._logger.info('failed to report {} to whiteboard'.format(FORKSERVER_EXECUTOR_STATS))

    def _send_one_keep_alive_req(self, cmd_type, timestamp):
        if timestamp - self._keep_alive_timestamp[cmd_type] < keep_alive_interval[cmd_type]:
            return

        # Send keep-alive check request, which is not limited by length of mapping dict.
        cmd_req = {
               CMD_SEQUENCE: self._cmd_seq,
               CMD_REQUEST_CHECK: True,
               CMD_REQUEST_TIMEOUT: FORK_SERVER_CMD_EXECUTE_DEF_SEC,
               CMD_REQUEST_TIMESTAMP: timestamp,
               CMD_TYPE: cmd_type
        }
        if not self._send(cmd_req, cmd_req):
            self._logger.error('Failed to send keep_alive request for type {}'
                               .format(cmd_req[CMD_TYPE]))
            return
        self._stats[KEEP_ALIVE_REQ] += 1
        self._logger.info('keep_alive req sent: {}'.format(cmd_req))
        # Update keep-alive timestamp
        self._keep_alive_timestamp[cmd_type] = timestamp


    def _send_keep_alive_req(self, timestamp):
        for cmd_type in range(CMD_TYPE_MAX):
            self._send_one_keep_alive_req(cmd_type, timestamp)

    def _keep_alive_failure(self, cmd_req):
        self._keep_alive_retry[cmd_req[CMD_TYPE]] -= 1
        # keep-alive check failed continuously, exit forkserver
        if not self._keep_alive_retry[cmd_req[CMD_TYPE]]:
            self._logger.warning('keep-alive check for type {} failed for {} times, exit'
                                 .format(cmd_req[CMD_TYPE], FORK_SERVER_KEEP_ALIVE_MAX_RETRY))
            self.set_exit()
        else:
            self._logger.warning('keep-alive retry left {} times'.format(self._keep_alive_retry[cmd_req[CMD_TYPE]]))

    def _receive_keep_alive_resp(self, cmd_req, cmd_resp):
        if not cmd_req.get(CMD_REQUEST_CHECK, False):
            return False
        self._stats[KEEP_ALIVE_RESP] += 1
        if not cmd_resp.get(CMD_RESPONSE_ERROR, None):
            self._logger.info('keep-alive check received resp {} for req{}'
                              .format(cmd_resp, cmd_req))
            self._keep_alive_retry[cmd_req[CMD_TYPE]] = FORK_SERVER_KEEP_ALIVE_MAX_RETRY
        else:
            # error response
            self._logger.warning('keep-alive check received error resp {} for req{}'
                                 .format(cmd_resp, cmd_req))
            self._keep_alive_failure(cmd_req)

        # report stats to data whiteboard
        self._report_stats(cmd_resp.get(CMD_RESPONSE_STATS, None))
        return True

    def _timeout_keep_alive_check(self, cmd_req):
        if not cmd_req.get(CMD_REQUEST_CHECK, False):
            return False
        self._logger.warning('keep-alive check timeout for req {}'.format(cmd_req))
        self._keep_alive_failure(cmd_req)

        # report stats to data whiteboard
        self._report_stats()
        return True


    def heartbeat_is_alive(self):
        timestamp = time.monotonic()
        for cmd_type in range(CMD_TYPE_MAX):
            if not self._keep_alive_retry[cmd_type] or \
               (self._keep_alive_timestamp[cmd_type] +
                keep_alive_interval[cmd_type] * self._keep_alive_retry[cmd_type] +
                FORK_SERVER_CMD_DELTA_SEC < timestamp):
                self._logger.info("Not send type[{}] heartbeat in {:.2f} seconds with {} retries left"
                                  .format(cmd_type, timestamp - self._keep_alive_timestamp[cmd_type],
                                          self._keep_alive_retry[cmd_type]))
                return False
        return True

    def _mapping_req_timeout(self, cmd_type, req, timestamp):
        # For requests have sent to forkserver, wait enough time to avoid false negative
        if cmd_type == CMD_TYPE_DEFAULT:
            # 3 delicated exectuors and the left 2 executors to hanle them when free.
            # At max 20(FORK_SERVER_EXCUTOR_MAX_COUNT * 4) requests in the queue.
            # In the worst case, we suppose that the default(pre-defined and limited latency)
            # commands shouldn't be blocked in pipe/queue for more than
            # FORK_SERVER_CMD_EXECUTE_DEFAULT_MAX_SEC * 3 + FORK_SERVER_CMD_DELTA_SEC.
            return req[CMD_REQUEST_TIMESTAMP] + req[CMD_REQUEST_TIMEOUT] + \
                   FORK_SERVER_CMD_EXECUTE_DEFAULT_MAX_SEC * 3 + FORK_SERVER_CMD_DELTA_SEC < timestamp
        else:
            # For heavy(long latency or custom) commands, 2 delicated excutors.
            # Use FORK_SERVER_CMD_EXECUTE_HEAVY_MAX_SEC *2 + FORK_SERVER_CMD_DELTA_SEC  as possible blocked time in pipe/queue
            return req[CMD_REQUEST_TIMESTAMP] + req[CMD_REQUEST_TIMEOUT] + \
                   FORK_SERVER_CMD_EXECUTE_HEAVY_MAX_SEC * 2 + FORK_SERVER_CMD_DELTA_SEC < timestamp


    def _pending_req_timeout(self, cmd_type, req, timestamp):
        # For request pending to wait for request, wait proper time to avoid blocking plugins too much when forkserver busy
        if cmd_type == CMD_TYPE_DEFAULT:
            # For default(pre-defined and limited latency)commands,
            # we suppose that they need be exectued in time, not pending for resource too long
            return req[CMD_REQUEST_TIMESTAMP] +  FORK_SERVER_CMD_PENDING_SEC < timestamp
        else:
            # For heavy(long latency or custom) commands, considering the exeution time is long, suppose that they
            # can be toleratable for more pending time: FORK_SERVER_CMD_EXECUTE_HEAVY_MAX_SEC + FORK_SERVER_CMD_PENDING_SEC
            return req[CMD_REQUEST_TIMESTAMP] + FORK_SERVER_CMD_EXECUTE_HEAVY_MAX_SEC + FORK_SERVER_CMD_PENDING_SEC < timestamp


    def _is_busy(self):
        for cmd_type in range(CMD_TYPE_MAX):
            if self._mapping_dicts[cmd_type] or self._pending_lists[cmd_type]:
                return True
        return False

    def _run(self):
        new_cmd_req = None
        while not self._exit_now:
            timestamp = (time.monotonic())
            # check forkserver process status
            if not self._process.is_alive():
                # notify main thread exit
                self.set_exit()
                break

            # keep-alive health check
            self._send_keep_alive_req(timestamp)

            # Get new request
            while (not self._req_queue.empty()) or new_cmd_req:
                self._logger.debug('try to get cmd_req, new_cmd_req:{}, req_queue empty is {}'.format(new_cmd_req, self._req_queue.empty()))
                if new_cmd_req:
                    cmd_req = new_cmd_req
                    new_cmd_req = None
                else:
                    cmd_req = self._req_queue.get()
                    self._stats[REQ] += 1
                    self._logger.debug('get cmd_req from queue {}'.format(cmd_req))
                if cmd_req[CMD_TYPE] >= CMD_TYPE_MAX:
                    self._stats[REQ_ERROR] += 1
                    self._logger.warning('Invalid req type: {}'.format(cmd_req))
                    cmd_req[CMD_REQUEST_RESP_QUEUE].put({CMD_RESPONSE_ERROR: FORK_SERVER_ERROR_INVALID_REQUEST})
                    continue
                # Put the req into the pending list
                self._pend_req(cmd_req, timestamp)

            # Send requests to forkserver
            for cmd_type in range(CMD_TYPE_MAX):
                while self._pending_lists[cmd_type] and self._has_resource(cmd_type):
                    cmd_req = self._pending_lists[cmd_type].pop(0)
                    if not self._send_req(cmd_req, timestamp):
                        # refresh timestamp and retry
                        timestamp = (time.monotonic())
                        if not self._send_req(cmd_req, timestamp):
                            # failed to send, return error to caller
                            cmd_req[CMD_REQUEST_RESP_QUEUE].put({CMD_RESPONSE_ERROR: FORK_SERVER_ERROR_SENT_FAILURE})
                            # refresh timestamp
                            timestamp = (time.monotonic())

            # Get response and return to caller
            while self._conn.poll():
                cmd_resp = conn_recv(self._conn, self._stats, self._logger, self)
                self._logger.debug('cmd_resp recvd: {}'.format(cmd_resp))
                if not cmd_resp:
                    # refresh timestamp
                    timestamp = (time.monotonic())
                    # skip to next loop to retry
                    continue
                if (not CMD_SEQUENCE in cmd_resp) or (not CMD_TYPE in cmd_resp):
                    self._logger.warning('Invalid resp, skip it: {}'.format(cmd_resp))
                    self._stats[RESP_ERROR] += 1
                    continue
                try:
                    cmd_req = self._mapping_dicts[cmd_resp[CMD_TYPE]].pop(cmd_resp[CMD_SEQUENCE])
                except KeyError:
                    self._logger.warning('Failed to found the req, which may have timed out. cmd_resp: {}'
                                         .format(cmd_resp))
                    self._stats[RESP_DROPPED_NO_REQ] += 1
                    continue
                # Handle the resp
                self._logger.debug('corresponding cmd_req found, have waited for {} seconds, cmd reach executor at: {}, execute_time:{}. {}'
                                   .format(timestamp - cmd_req[CMD_REQUEST_TIMESTAMP], cmd_resp.get(CMD_REQUEST_TIMESTAMP, 0),
                                           cmd_resp.get(CMD_RESPONSE_EXECUTE_TIME, 0), cmd_req))
                self._stats[RESP_RCVD] += 1
                # Check keep-alive response firstly, and then send other resp back to the caller
                if not self._receive_keep_alive_resp(cmd_req, cmd_resp):
                    if cmd_req.get(CMD_REQUEST_TIMED_OUT, False):
                        cmd_resp[CMD_RESPONSE_ERROR] = FORK_SERVER_ERROR_EXECUTE_TIMEOUT
                    cmd_req[CMD_REQUEST_RESP_QUEUE].put(cmd_resp)
                    self._stats[RESP] += 1
                    self._logger.debug('put resp {} into queue {}'
                                      .format(cmd_resp, cmd_req[CMD_REQUEST_RESP_QUEUE]))

            # Handle timeout requests per-second
            if timestamp >= self._timeout_check_timestamp + 1:
                self._logger.debug('check timeout requests')
                self._timeout_check_timestamp = timestamp
                # Mappling requests timeout
                for cmd_type in range(CMD_TYPE_MAX):
                    for key in list(self._mapping_dicts[cmd_type].keys()):
                        req = self._mapping_dicts[cmd_type][key]
                        if req.get(CMD_REQUEST_TIMED_OUT, False) or not self._mapping_req_timeout(cmd_type, req, timestamp):
                            continue
                        self._logger.info('mapping req timeout {}'.format(req))
                        # Check keep-alive check req
                        self._timeout_keep_alive_check(req)
                        # mark it as timeout but not to remove
                        req[CMD_REQUEST_TIMED_OUT] = True
                        self._stats[REQ_DROP_TIMEOUT] += 1
                # Pending requests timeout
                # requests in the list is in the order of inserting, check the first ones till one not timeout.
                for cmd_type in range(CMD_TYPE_MAX):
                    while len(self._pending_lists[cmd_type]):
                        req = self._pending_lists[cmd_type][0]
                        if not self._pending_req_timeout(cmd_type, req, timestamp):
                            break
                        self._logger.info('pending req {} timeout'.format(req))
                        req[CMD_REQUEST_RESP_QUEUE].put({CMD_RESPONSE_ERROR: FORK_SERVER_ERROR_PENDING_TIMEOUT})
                        del self._pending_lists[cmd_type][0]
                        self._stats[REQ_DROP_NO_RESOURCE] += 1

            # Wait for new response back or new request coming
            if (self._is_busy()):
                try:
                    self._conn.poll(timeout=FORK_SERVER_BUSY_CHECK_INTERVAL)
                except OverflowError as e:
                    import traceback
                    trace = traceback.format_exc()
                    logger.info("conn poll OverflowError, maybe due to system clock change: {}, {}".format(e, trace))
                    self.set_exit(force=True)
            else:
                try:
                    new_cmd_req = self._req_queue.get(timeout = FORK_SERVER_IDLE_CHECK_INTERVAL)
                    self._stats[REQ] += 1
                    self._logger.debug('get new_cmd_req from queue {}'.format(new_cmd_req))
                except Empty:
                    pass

        self._logger.info('forkserver monitor stopped.')
        self._logger.info('forkserver monitor stats: {}'.format(self._stats))
        self._logger.info('ForkServer stats: {}'.format(ForkServer._stats))

    def run(self):
        self._logger.info('monitor thread for forkserver started')
        try:
            self._run()
        except Exception as e:
            import traceback
            trace = traceback.format_exc()
            self._logger.info("failed: {}, {}".format(e, trace))

    def stop(self):
        if self._process.is_alive() :
            # terminate forkserver process
            self._logger.info("stopping forkserver process: {}".format(self._process))
            # For linux, send terminal signal to forkserver process.
            # For windows, it will call corresponding API to terminate the process directly.
            self._process.terminate()

        if self.is_alive() or not self._exit_now:
            # Stop the monitor thread and notify main thread
            self._logger.info("stopping forkserver monitor")
            self.set_exit()

    def set_exit(self, force=False):
        if not self._exit_now:
            self._logger.info('set exit_now flag of forkserver monitor')
            self._exit_now = True
        if force:
            self._logger.warning("Force exit immediately")
            os._exit(0)

    @property
    def exit_now(self):
        return self._exit_now

    def wait_stop(self):
        # wait for forkserver process exit
        self._process.join(10)
        if self._process.is_alive():
            self._logger.warning('forkserver process cannot exit in 10 secs, kill it forcelly!')
            try:
                if sha.preconfig.SYSTEM != sha.preconfig.PLATFORM_WINDOWS:
                    os.kill(self._process.pid, signal.SIGKILL)
                    self._process.join()
                else:
                    # On windows, the forkserver process should have been terminated directly,
                    # No need to kill again.
                    self._logger.error('forkserver process on windows not terminated yet!')

                self._logger.info("forkserver process exited: {}".format(self._process))
            except Exception as e:
                import traceback
                trace = traceback.format_exc()
                self._logger.info("failed: {}, {}".format(e, trace))
        else:
            self._logger.info("forkserver process exited: {}".format(self._process))

        # Wait for forksever monitor exit
        self.join(10)
        if self.is_alive():
            self._logger.info('forkserver monitor cannot exit in 10 secs, exit forcelly!')


class ForkServer:
    _logger = sha.preconfig.get_sub_logger(__name__)
    _parent_conn = None
    _child_conn = None
    _server_process = None
    _server_monitor = None
    _req_queue = None
    _stats = {
       REQ: 0,
       RESP: 0,
       RESP_ERROR: 0
    }

    @classmethod
    def start(cls):
        cls._logger.info('starting process of forkserver')
        cls._parent_conn, cls._child_conn = Pipe()
        cls._server_process = Process(target=fork_server, args=(cls._child_conn,))
        cls._server_process.start()
        # prepare request queue
        cls._req_queue = Queue(FOEK_SERVER_REQUEST_QUEUE_SIZE)
        # Create a new thread to monitor the fork sever process
        cls._server_monitor = ForkServerMonitor(cls._server_process, cls._parent_conn, cls._req_queue)
        cls._server_monitor.daemon = True
        cls._server_monitor.start()

        return cls._server_monitor, cls._server_process.pid


    @classmethod
    def _classify_cmd(cls, timeout, custom):
        timeout_sec =  FORK_SERVER_CMD_EXECUTE_DEF_SEC if timeout is None else timeout
        # request timeout shouldn't be too long
        if timeout_sec > FORK_SERVER_CMD_EXECUTE_HEAVY_MAX_SEC:
            raise Exception("Invalid command request with too long timeout: {}".format(timeout))

        if custom or timeout_sec > FORK_SERVER_CMD_EXECUTE_DEFAULT_MAX_SEC:
            req_type = CMD_TYPE_HEAVY
        else:
            req_type = CMD_TYPE_DEFAULT

        return req_type, timeout_sec

    @classmethod
    def check_output(cls, cmd, input=None, shell=False, timeout=FORK_SERVER_CMD_EXECUTE_DEF_SEC, custom=False, env=None):
        req_type, timeout_sec = cls._classify_cmd(timeout, custom)
        resp_queue = Queue(1)
        cmd_req = {
            CMD_REQUEST_COMMAND: cmd,
            CMD_REQUEST_INPUT: input,
            CMD_REQUEST_SHELL: shell,
            CMD_REQUEST_TIMEOUT: timeout_sec,
            CMD_REQUEST_RESP_QUEUE: resp_queue,
            CMD_REQUEST_ENV: env,
            CMD_TYPE: req_type
        }

        start = time.monotonic()
        cls._req_queue.put(cmd_req)
        cls._stats[REQ] += 1
        cls._logger.debug("Request to run command: {}".format(cmd_req))
        # Due to some commands may hanging, not to check its own execution timeout, but just check
        # whether forkserver exit or the keep-alive heartbeat stopped(for the case fork monitor thread crashed)
        cmd_resp = None
        while cls._server_monitor.heartbeat_is_alive() and not cls._server_monitor.exit_now:
            try:
                cmd_resp = resp_queue.get(timeout=1)
                break
            except Empty:
                pass
        cls._logger.debug("Total execute_time: {} seconds for req{}".format(time.monotonic()-start, cmd_req))

        if not cmd_resp:
            if not cls._server_monitor.exit_now:
                error = "fork server timeout"
                # Forkserver monitor hang or crashed, trigger SHA exit
                cls._server_monitor.set_exit()
            else:
                error = "fork server exit"
        else:
            # Check error returned by forkserver
            error = cmd_resp.get(CMD_RESPONSE_ERROR, None)
            if error:
                cls._stats[RESP_ERROR] += 1

        if error:
            cls._logger.warning("Failed to run command: {} with error {}".format(cmd_req, error))
            raise Exception("Failed to run command: {}".format(error))

        cls._stats[RESP] += 1

        # Check catched exception
        e = cmd_resp.get(CMD_RESPONSE_EXCEPTION, None)
        if e:
            raise e

        return cmd_resp[CMD_RESPONSE_OUTPUT]
